Pro esi_flatfixer,flatlist,wavearr,slitarr,orderlist,combflat,combflatvar,$
                  bkspace=bkspace,arc=arc,refarc=refarc,scattered=scattered
                   
;+-------------------------------------------------------------------
;
; ESI_FLATFIXER    9/2010
;
; Median-combine and normalize ESI flats.
;
; Input flats are median combined, the combined flat is normalized,
; and the formal variance in the normalized flat is computed assuming
; Poisson statistics.
;
; Assumes 1x1 binning.
;
; INPUTS
;       flatlist  - String, name of single-column ASCII file containing
;                   the FITS filenames of flats to be combined, 
;                   OR string array of FITS filenames of flats to be
;                   combined.
;       wavearr   - 2D array or fits filename, where the pixel values
;                   are uncalibrated wavelengths (in pixels).
;       slitarr   - 2D array or fits filename, where the pixel values
;                   are position along the slit.
;       orderlist - Name of ASCII file containing the order
;                   information [CCD, order #, center, left edge,
;                   right edge], as output by esi_slitgrid.pro.
;
; KEYWORDS
;       bkspace   - Spacing of bspline break points, in pixels (or
;                   whatever units are used for the input wavearr).
;                   Defaults are specified above for internal and dome
;                   flats.  Smaller spacing may be needed for
;                   exposures that contain emission lines, but larger
;                   spacing should be used to preserve fringe
;                   features.
;       arc       - Arc exposure taken contemporaneously with flats
;       refarc    - Arc exposure taken with exposures used to create
;                   wavearr and slitarr.  If both arc and refarc
;                   keywords are fit, wavearr and slitarr will be registered
;                   to the flatfield exposures prior to fitting.
;       scattered - Subtract scattered light by fitting a 2D polynomial
;                   to the inter-order counts.
;
;  OUTPUTS
;       combflat    - Median-combined, normalized flat field.  May
;                     either be a variable or a fits filename to which
;                     array is to be written.
;       combflatvar - Formal variance in the combined, normlized flat.
;                     May either be a variable or a fits filename to
;                     which array is to be written.
;
; HISTORY
;       Adapted from MAGE_FLATFIXER 9/26/2010 GDB
;---------------------------------------------------------------------

if (n_params() eq 0) then begin
   print,'CALLING SEQUENCE:  '
   print,' esi_flatfixer,flatlist,wavearr,slitarr,orderlist,combflat,'
   print,'               combflatvar,bkspace=bkspace,arc=arc,refarc=refarc,'
   print,'               /scattered'
   return
endif

;;; Read in order information
order_templ =  {   $
          version: 1.0, $
          datastart: 0, $
          delimiter: ' ', $        ; Fields separated by white space
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 5, $
          fieldtypes: [7,4,4,4,4], $
          fieldnames: ['ccd','ordernum','center','left','right'], $
          fieldlocations: [2,11,18,30,42], $
          fieldgroups: [0,1,2,3,4] $
        }
orderinfo       = read_ascii(orderlist,template=order_templ)
tmp_ordernum    = orderinfo.ordernum
tmp_ordercenter = orderinfo.center
tmp_orderedge1  = orderinfo.left
tmp_orderedge2  = orderinfo.right
; Sort orders in order of increasing slit position
order_order = sort(tmp_ordercenter)
ordernum    = tmp_ordernum(order_order)
ordercenter = tmp_ordercenter(order_order)
orderedge1  = tmp_orderedge1(order_order)
orderedge2  = tmp_orderedge2(order_order)
n_orders    = n_elements(ordercenter)

;;; Read in flat filenames
flatlist_sz = size(flatlist)
if (flatlist_sz(0) eq 0) then begin
   ; ASCII file with names of flat fields
   flat_templ = {   $
          version: 1.0, $
          datastart: 0L, $
          delimiter: ' ', $
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 1L, $
          fieldtypes: [7L], $
          fieldnames: ['name'], $
          fieldlocations: [6L], $
          fieldgroups: [0L] $
        }
   list     = read_ascii(flatlist,template=flat_templ)
   flatfile = list.name
endif else begin
   ; Array with names of flat fields
   flatfile = flatlist
endelse
nflat    = n_elements(flatfile)

;;; Read in wave and slit grids.

if (size(wavearr,/tname) eq 'STRING') then begin
   checkexists = findfile(wavearr,count=nfiles)
   if (nfiles eq 0) then begin
      print,'Wavelength file not found.'
      return
   endif else begin
      wavegrid = readfits(wavearr)
   endelse
endif else wavegrid = wavearr

if (size(slitarr,/tname) eq 'STRING') then begin
   checkexists = findfile(slitarr,count=nfiles)
   if (nfiles eq 0) then begin
      print,'Slit position file not found.'
      return
   endif else begin
      slitgrid = readfits(slitarr)
   endelse
endif else slitgrid = slitarr

; Arrays of x and y-values
sz   = size(wavegrid)
xsz  = sz(1)
ysz  = sz(2)
xarr = rebin(lindgen(xsz),xsz,ysz)
yarr = rebin(rotate(lindgen(ysz),1),xsz,ysz)

;;; Optional - Register wave and slit grids to flat field exposures

if (keyword_set(arc) and keyword_set(refarc)) then begin
   print,'Registering slit and wave grids to '+arc
   flat_arc = esi_subbias(readfits(arc))
   grid_arc = esi_subbias(readfits(refarc))
   esi_register,grid_arc,flat_arc,p,q
   print,'POLY_2D terms:'
   print,p,q
   ; Suppress pixels that were interpolated from order edges
   ordergrid = 1 + 0.*slitgrid
   interorder_ls = where(slitgrid eq -1000)
   ordergrid(interorder_ls) = 0.0   
   newordergrid = poly_2d(ordergrid,p,q,1,missing=0.0)
   slitgrid     = poly_2d(slitgrid,p,q,1,missing=-1000)
   wavegrid     = poly_2d(wavegrid,p,q,1,missing=-1000)
   suppress_ls  = where(newordergrid lt 1,n_suppress)
   if (n_suppress ne 0) then begin
      slitgrid(suppress_ls) = -1000
      wavegrid(suppress_ls) = -1000
   endif
endif

;;;
;;; Median combine flats
;;;

; Arrays to store processed flats.
bigflat    = fltarr(xsz,ysz,nflat)
texp       = fltarr(nflat)         ; Exposure times

; Read in flats
for k=0,nflat-1 do begin
   tmp_flatfile = flatfile(k)
   checkexists  = findfile(tmp_flatfile,count=nfiles)
   if (nfiles eq 0) then begin
      print,'Flat file '+tmp_flatfile+' not found.'
      return
   endif else begin
      tmp_flat       = readfits(tmp_flatfile,flat_hdr)
      ; Subtract bias
      bigflat(*,*,k) = esi_subbias(tmp_flat)
      texp(k)        = sxpar(flat_hdr,'ELAPTIME')
   endelse
endfor

; Determine gain and read noise
gain_name = strtrim(sxpar(flat_hdr,'CCDGAIN'),2)
case gain_name of
   'low' : begin
              gn  = 1.29  ; e-/DN  Avergage gain
              rdn = 2.7   ; e-
           end
   'high': begin
              gn  = 0.5
              rdn = 2.7
           end
   else:   begin
              print,'Unknown gain'
              return
           end
endcase               

; Check that all flats have similar exposure times
texp_fracdiff = abs(texp - mean(texp))/mean(texp)
if (max(texp_fracdiff) gt 0.01) then begin
   print,'ESI_FLATFIXER: Different exposure times detected - sky flats?'
   print,'                Scaling counts to match.'
   print,'                Note: Formal variance in combined flat may not'
   print,'                be correct.'
   temp_medflat = median(bigflat,dim=3,/even)
   ls           = where(temp_medflat ge 300)
   med_level    = median(temp_medflat(ls))
   for i=0,nflat-1 do $
      bigflat(*,*,i) = bigflat(*,*,i) * med_level / median(bigflat(*,*,i))
endif

; Median combine
if (nflat gt 1) then begin
   medflat = median(bigflat,dim=3,/even)
endif else begin
   medflat = bigflat
endelse

; Remove difference in gain.  Adjust counts up and down about the mean
; in order to minimize problems when identifying bad pixels based on
; the normalized flat field in longslit_reduce.
gain_ratio           = 0.93  ; gain2(rhs) / gain1(lhs)
gain_factor2         = 0.5*(1 + gain_ratio)
gain_factor1         = gain_factor2 / gain_ratio
medflat(24:1047,*)   = medflat(24:1047,*)   * gain_factor1
medflat(1048:2071,*) = medflat(1048:2071,*) * gain_factor2

; Optional - Subtract scattered light
if keyword_set(scattered) then begin
   print,'Subtracting scattered light...'
   ; Select regions in between orders
   maskval    = -100
   interorder = maskval + 0.*medflat
   for i=0,n_orders-2 do begin
      inter_ls = where(slitgrid ge orderedge2(i)+5   and $
                       slitgrid le orderedge1(i+1)-5 and $
                       xarr ge 24 and xarr le 2071,n_inter)
      if (n_inter ne 0) then interorder(inter_ls) = medflat(inter_ls)
   endfor
   ; Include region below lowest order and a above highest order
   bottom_ls = where(slitgrid le orderedge1(0)-5 and $
                     xarr ge 24,n_bottom)
   if (n_bottom ne 0) then interorder(bottom_ls) = medflat(bottom_ls)
   top_ls    = where(slitgrid ge orderedge2(n_orders-1)+5 and $
                     xarr le 2071,n_top)
   if (n_top ne 0) then interorder(top_ls) = medflat(top_ls)
   ; Mask non-illuminated regions
   nonillum_ls = where(xarr lt 24 or xarr gt 2071,n_nonillum)  ; 1x1 binning
   if (n_nonillum ne 0) then interorder(nonillum_ls) = maskval
   ; Mask saturated regions.  Choose a low count threshhold, since
   ; scattered light should be relatively faint.
   saturated_ls = where(medflat ge 40000,n_saturated)
   if (n_saturated ne 0) then interorder(saturated_ls) = maskval
   ; Fit 2D polynomial
   fit_ls = where(interorder ne maskval,n_fit)
   if (n_fit eq 0) then begin
      print,'ESI_FLATFIXER: No good pixels for fitting scattered light.'
      return
   endif
   interorder_data = fltarr(3,n_fit)
   interorder_data(0:1,*) = array_indices(interorder,fit_ls)
   interorder_data(2,*)   = interorder(fit_ls)
   fit_order = 4
   temp_fit  = sfit(interorder_data,fit_order,/irregular,kx=fitterms)
   ; Evaluate polynomial over entire array and subtract
   scattered_light = 0.*medflat
   for i=0,fit_order do begin
      for j=0,fit_order do begin
         scattered_light = scattered_light + fitterms(j,i)*(xarr^i)*(yarr^j)
      endfor
   endfor
   if (n_nonillum ne 0) then scattered_light(nonillum_ls) = 0.
   medflat = medflat - scattered_light
endif

;;;
;;; Normalize
;;;

; Set break point spacing for b-spline fit.
if keyword_set(bkspace) then bk_space=bkspace else bk_space=100.

; Arrays of bad pixels/rows/columns.  Flag with 1.
mask = lonarr(xsz,ysz)
mask(64:161,3858:3932)   = 1
mask(446:463,2648:ysz-1) = 1
mask(561,2563:3370)      = 1
mask(903:928,3811:4037)  = 1

; Output arrays
normflat     =  1 + fltarr(xsz,ysz)
normflat_var = -1 + fltarr(xsz,ysz)

; Array of edges, useful for non-diffused flats.  For faint orders
; pixels in the center of the array can look like edges using the sobel
; method.  Therefore, construct a median "edge value" as a function
; of slit position for selecting pixels away from the edges.
sm_flat    = medflat(24:2071,1400:1900)
sm_slit    = slitgrid(24:2071,1400:1900)
edge_array = sobel(sm_flat)/float(sm_flat)
h          = histogram(sm_slit,bin=0.5,locations=slit_bin,rev=ri)
slit_bin   = slit_bin + 0.25
edge_bin   = 0.*slit_bin
for i=0,n_elements(h)-2 do if (h(i) ne 0) then $
   edge_bin(i) = median(edge_array(ri(ri(i):ri(i+1)-1)))

; Meausre a "minimum" variance from a small region in the upper
; righ-hand corner
min_sdev = robust_sigma(medflat(1960:2060,4000:4090))
min_var  = min_sdev^2

;;; For each order, identify illuminated region and fit.

for i=0,n_orders-1 do begin

   print,' '
   print,'Order '+strtrim(ordernum(i),2)

   ; Identify illumated region
   slit_lo  = orderedge1(i)
   slit_hi  = orderedge2(i)
   illum_ls = where(slitgrid ge slit_lo and slitgrid lt slit_hi and $
                    xarr ge 24 and xarr le 2071 and wavegrid gt -500,n_illum)
   
   ; Identify region to be fit.  Avoid edge regions.
   ; For ESI, avoid left edge of the slit, where signal descreases
   ; in some flats.
   edge_lo = min(slit_bin(where(slit_bin ge slit_lo and edge_bin lt 1.0)))
   edge_hi = max(slit_bin(where(slit_bin le slit_hi and edge_bin lt 1.0)))
   fit_ls  = where(slitgrid ge ((slit_lo+25) > edge_lo) and $
                   slitgrid lt (slit_hi < edge_hi) and $
                   xarr ge 24 and xarr le 2071     and $
                   mask ne 1 and wavegrid gt -500,n_fit)
   fitslit = slitgrid(fit_ls)
   fitwave = wavegrid(fit_ls)
   fitflat = medflat(fit_ls)

   ; Assume Poisson noise (If there are two flats then they will
   ; be averaged, otherwise they are median combined.)
   if (nflat gt 2) then med_factor = !pi/2. else med_factor = 1
   fitflat_var = med_factor*(fitflat/gn + (rdn/gn)^2)/nflat
   fitflat_var = fitflat_var > min_var
   fitflat_invvar = 1./fitflat_var
   
   ; Identify potential bad pixels before fitting.
   print,'Pre-filtering bad pixels...'
   maxwave = max(fitwave)
   minwave = min(fitwave)
   first_filter = 1
   filter_sz = 2*bk_space
   n_filters = ceil((maxwave-minwave)/filter_sz)
   ; Default is not to filter any points
   filter_ls    = -1
   first_filter = 1
   h = histogram(fitwave,bin=2*bk_space,locations=wave_bin,rev=ri)
   for j=0,n_elements(h)-1 do begin
      if (h(j) lt 10) then continue
      tmp_ls     = ri(ri(j):ri(j+1)-1)
      tmp_wave   = fitwave(tmp_ls)
      tmp_flat   = fitflat(tmp_ls)
      ; Fit a low-order polynomial to this region, and mask
      ; outliers from the fit
      tmp_pfit   = robust_poly_fit(tmp_wave,tmp_flat,1,tmp_fit)
      tmp_sdev   = robust_sigma(tmp_flat-tmp_fit)
      tmp_filter_ls = where(abs(tmp_flat-tmp_fit) ge 4.*tmp_sdev,tmp_n_filter)
      if (tmp_n_filter ne 0) then begin
         if first_filter then begin
            filter_ls    = tmp_ls(tmp_filter_ls)
            first_filter = 0   
         endif else begin
            filter_ls    = [filter_ls,tmp_ls(tmp_filter_ls)]
         endelse
      endif
   endfor
   if (total(filter_ls) ne -1) then fitflat_invvar(filter_ls) = 0

   ; Perform b-spline fit in the dispersion direction.  Include a
   ; variable linear correction to the illumination function.
;   sset = bspline_iterfit(fitwave,fitflat,invvar=fitflat_invvar,$
;                          bkspace=bk_space,yfit=fitflat_model,upper=5,lower=5,$
;                          x2=fitslit,illum=2)
   sset = bspline_iterfit(fitwave,fitflat,invvar=fitflat_invvar,$
                          bkspace=bk_space,yfit=fitflat_model,upper=5,lower=5)

   ; Evaluate bspline fit over illuminated region and divide out
   medflat_model = fltarr(xsz,ysz)
;   medflat_model(illum_ls) = bspline_valu(wavegrid(illum_ls),sset,$
;                                          x2=slitgrid(illum_ls))
   medflat_model(illum_ls) = bspline_valu(wavegrid(illum_ls),sset)
   normflat(illum_ls) = medflat(illum_ls) / medflat_model(illum_ls)
   
   ; Estimate variance in the normalized flat.  Non-illuminated regions 
   ; are flagged with -1.
   normflat_var(illum_ls) = $
         (med_factor*(medflat_model(illum_ls)/gn + (rdn/gn)^2)/nflat) / $
         (double(medflat_model(illum_ls))^2)

   ;; Fit residual illumination function.
   ;illumfit_ls = where(slitgrid ge slit_lo and slitgrid lt slit_hi and $
   ;                    xarr ge 24 and xarr le 2071 and mask ne 1 and $
   ;                    abs(normflat-1) le 0.5,n_illum)
   ;
   ;illum_slit       = slitgrid(illumfit_ls)
   ;illum_flat       = normflat(illumfit_ls)
   ;illumflat_invvar = 1./normflat_var(illumfit_ls)
   ;illum_sset = bspline_iterfit(illum_slit,illum_flat,invvar=illumflat_invvar,$
   ;                             bkspace=1.0,yfit=illum_fit,upper=5,lower=5)
   ;
   ;; Evalutate illumination function over entire slit and divide out
   ;full_illum_fit   = bspline_valu(slitgrid(illum_ls),illum_sset)
   ;
   ;normflat(illum_ls)     = normflat(illum_ls) / full_illum_fit
   ;normflat_var(illum_ls) = normflat_var(illum_ls) / (full_illum_fit^2)

endfor

; Restore difference in gain.
normflat(24:1047,*)       = normflat(24:1047,*)   / gain_factor1
normflat(1048:2071,*)     = normflat(1048:2071,*) / gain_factor2
normflat_var(24:1047,*)   = normflat_var(24:1047,*)   / (gain_factor1^2)
normflat_var(1048:2071,*) = normflat_var(1048:2071,*) / (gain_factor2^2)

;;;
;;; Set output arrays or write fits files
;;;

hnormflat = flat_hdr
sxaddhist,'-------------------------------------------------------',hnormflat
sxaddhist,'Flatfield processing done using ESI_FLATFIXER',hnormflat
sxaddhist,'   Flatfield frames combined:',hnormflat
for i=0,nflat-1 do $
   sxaddhist,'      '+flatfile(i),hnormflat
sxaddhist,'-------------------------------------------------------',hnormflat

sxaddpar,hnormflat,'BZERO',0.0

if (size(combflat,/tname) eq 'STRING') then begin
   print,'Writing '+combflat
   writefits,combflat,normflat,hnormflat
endif else combflat = normflat

if (size(combflatvar,/tname) eq 'STRING') then begin
   print,'Writing '+combflatvar
   writefits,combflatvar,normflat_var,hnormflat
endif else combflatvar = normflat_var

print,'Done.'

return
end
