Pro esi_register,im1,im2,p,q,nx=nx,ny=ny,offset=offset,nline=nline,$
                  coord1=coord1,coord2=coord2

;+-----------------------------------------------------------------------
;
; ESI_REGISTER
;
; Use arc line or sky line centroids to find the best polynomial warping
; that matches two 2D ESI spectra.
;
; Performs a surface fitting of x,y coordinates in im1 as a function of
; the corresponding coordinates in im2.  Returns arrays of polynomial
; coefficients that can be input into POLY_2D to create a version of
; im1 matched to im2, e.g.:
;        
;           im1_match = poly_2d(im1,p,q,1,missing=0)
;
; NOTE: The /max_degree is set for SFIT below.  May wish to change this
; if more complicated transformations are needed, e.g., cases in which
; there is a large shift in both x and y.
;
; If registering arc lamp exposures, the resulting difference in the
; positions of the arc lines should be nearly imperceptible.  
;
; NOTE: *** It is essential that im1 and im2 be bias-subtracted. ***
;
; INPUTS
;        im1      - Array, 2D echellogram to be shifted
;        im2      - Array, 2D echellogram, same size as im1, to which im1
;                   is to be matched
;
; KEYWORDS
;        nx       - Max polynomial order for fitting x coordinates.
;                   The /max_degree keyword is set for SFIT below, so
;                   this is the maximum degree for all dimensions
;                   combined.  Default is 1, which is good for small
;                   shifts (~< 10 pixels).  Setting order >= 2 is
;                   recommended for large shifts.
;        ny       - Max polynomial order for fitting y coordinates.
;                   The /max_degree keyword is set for SFIT below, so
;                   this is the maximum degree for all dimensions
;                   combined.  Default is 1, which is good for small
;                   shifts (~< 10).  Setting order >= 2 is recommended
;                   for large shifts.
;        offset   - Two-element vector [dx,dy] giving the rough
;                   offset from a pixel on im1 to the corresponding
;                   pixel on im2.  Default is to iteratively search
;                   for the best offset between frames.
;        nline    - Number of lines to attempt to identify on each
;                   spectrum.  Default is 40. 
;        coord1   - Coordinates of the lower left-hand and upper 
;                   right-hand corners of im1.  If im1 is an m x n
;                   array, then coord1 = [0,m-1,0,n-1].
;        coord2   - Coordinates on im2 that correspond phsyically
;                   to coord1 on im1.  coord1 and coord2 can be
;                   input into im_rot, which will give the correct
;                   transformation in the case of a linear fit,
;                   although it is easier to use ploy_2d with the
;                   p and q parameters.
;
; OUTPUTS
;        p,q      - Arrays of polynomial coefficients that give the
;                   coordinates of im1 as a function of the matching
;                   coordinates in im2.  The can be input into POLY_2D
;                   to create a version of im1 registered to im2:
;                   See POLY_2D for definitions of p and q.
; 
; HISTORY
;        Adapted from XSHOOTER_REGISTER 10/1/2010 GDB
;------------------------------------------------------------------------

if (n_params() lt 2) then begin
   print,'CALLING SEQUENCE:'
   print,'     esi_register,im1,im2,p,q,nx=nx,ny=ny,offset=offset,'
   print,'                   nline=nline,coord1=coord1,coord2=coord2'
   return
endif

;;; Swap x and y such that lines run along columns (this is leftover
;;; from how the routine was implemented for NIRSPEC and MagE.)
flip_im1 = rotate(im1,4)
flip_im2 = rotate(im2,4)

;;; Image size
sz  = size(flip_im1)
xsz = sz(1)
ysz = sz(2)
xarr = dblarr(xsz,ysz)
yarr = dblarr(xsz,ysz)
for i=0,xsz-1 do xarr(i,*) = i
for i=0,ysz-1 do yarr(*,i) = i
illum_xsz = xsz
illum_ysz = ysz

;;; Median filter to remove cosmic rays (probably not necessary for
;;; arcs)
filter_length = 11
medim1        = 0.*flip_im1
medim2        = 0.*flip_im2
; Lines run along coulmns.  Must rotate before filtering.
for i=0,xsz-1 do begin
   slice = rotate(flip_im1(i,*),1)
   filt  = median(slice,filter_length)
   medim1(i,*) = rotate(filt,3)
endfor
for i=0,xsz-1 do begin
   slice = rotate(flip_im2(i,*),1)
   filt  = median(slice,filter_length)
   medim2(i,*) = rotate(filt,3)
endfor

;;; Find skyline centroids
if keyword_set(nline) then n_line = nline $
   else n_line = 40      ; Max number of lines centroids to find
max_attempt   = n_line*4 ; Max number of attempts to locate lines
top_border    = 50       ; Border at edge of image to avoid
bottom_border = 50 
left_border   = 50 
right_border  = 50 
satlevel=6e4  ; Saturation level
if (xsz lt 3000) then specbin=2. else specbin=1.  ; Flipped x and y!
min_width     = 1.0 / specbin
max_width     = 8.0 / specbin
for k=1,2 do begin
   if (k eq 1) then image = medim1 else image = medim2
   ; Suppress borders
   image(0:left_border-1,*)        = 0.
   image(illum_xsz-right_border:illum_xsz-1,*) = 0.
   image(*,0:bottom_border-1)      = 0.
   image(*,illum_ysz-top_border:illum_ysz-1)   = 0.
   ; Suppress saturated regions
   bad_ls = where(image ge satlevel,n_bad)
   if (n_bad ne 0) then begin
      bad_indices = array_indices(image,bad_ls)
      xbad        = bad_indices(0,*)
      ybad        = bad_indices(1,*)
      xlo         = (xbad-30) > 0
      xhi         = (xbad+30) < (xsz-1)
      ylo         = (ybad-30) > 0
      yhi         = (ybad+30) < (ysz-1)
      for i=0L,long(n_bad-1) do image(xlo(i):xhi(i),ylo(i):yhi(i)) = 0.
   endif
   xlist = dblarr(n_line)
   ylist = dblarr(n_line)
   n_found   = 0
   n_attempt = 0
   ; Compute image histogram
   h    = histogram(image,rev=ri)
   n_ri = n_elements(ri)
   j    = 0L
   while (n_found lt n_line and n_attempt lt max_attempt) do begin
      n_attempt = n_attempt + 1
      ; Find brightest remaining pixel
      pixel_found = 0
      j = j+1
      while not(pixel_found) do begin
         i = ri(n_ri-j)
         if (image(i) ne 0) then pixel_found=1 else j=j+1
      endwhile
      ; Peak up on this location
      ind    = array_indices(image,i)
      x      = ind(0)
      y      = ind(1)
      tol    = 0.1
      dx     = 10
      dy     = 10
         xside  = 10. / specbin  ; > 1/2 line width to include some of the dark level
         yside  = 80.            ; Should be smaller than border, > 1/2 arc line length
      count  = 0
      while (abs(dx) gt tol or abs(dy) gt tol and count le 100 and $
             x ge xside and x le xsz-1-xside and $
             y ge yside and y le ysz-1-yside) do begin
         ;Compute flux-weighted centroid
         xcut  = xarr(x-xside:x+xside,y-yside:y+yside)
         ycut  = yarr(x-xside:x+xside,y-yside:y+yside)
         imcut = image(x-xside:x+xside,y-yside:y+yside)
         xcent = total(xcut*imcut)/total(imcut)
         ycent = total(ycut*imcut)/total(imcut)
         dx    = xcent - x
         dy    = ycent - y
         x     = (xside+5) > (xcent < (xsz-xside-5))
         y     = (yside+5) > (ycent < (ysz-yside-5))
         count = count + 1
      endwhile
      ; Count this line only if it does not contain saturated
      ; pixels, if it has a fitted width consistent with a 
      ; single isolated line, and if it not on the edge.
      if (max(imcut) lt satlevel and $
          (x ge xside+left_border and x le xsz-1-xside-right_border) and $
          (y ge yside+bottom_border and y le ysz-1-yside-top_border)) $
        then begin     
         ; Fit a Gaussian in x
         med_imcut = median(image(x-xside:x+xside,y-1:y+1),dim=2)
         gfit = gaussfit(xarr(x-xside:x+xside,y),med_imcut,$
                         a,nterms=4)  ; 4 terms: NIRSPEC has high dark current
         max_imcut = max(med_imcut)
         ; Note: Eliminated height criterion from SPEC_REGISTER
         if (a(2) ge min_width and a(2) le max_width) then begin
            if (n_found eq 0) then begin
               xlist = [a(1)]
               ylist = [y]
            endif else begin
               xlist = [xlist,a(1)]
               ylist = [ylist,y]
            endelse
            n_found = n_found + 1
         endif
      endif
      ; Suppress counts in the vicinity of the current line.
      ; If x and y are non-sensible, then supress counts around
      ; the orginial position.
      if (finite(x) eq 1 and finite(y) eq 1) then begin
         xlo = (x-xside) > 0
         xhi = (x+xside) < (xsz-1)
         ylo = (y-yside) > 0
         yhi = (y+yside) < (ysz-1)
      endif else begin
         xlo = (ind(0)-xside) > 0
         xhi = (ind(0)+xside) < (xsz-1)
         ylo = (ind(1)-yside) > 0
         yhi = (ind(1)+yside) < (ysz-1)
      endelse
      image(xlo:xhi,ylo:yhi) = 0
   endwhile
   if (k eq 1) then begin
      xlist1 = xlist
      ylist1 = ylist
      n_line1 = n_elements(xlist1)  ; Number of lines found
   endif else begin
      xlist2 = xlist
      ylist2 = ylist
      n_line2 = n_elements(xlist2)
   endelse
endfor

;;; Get rough offset between frames
if keyword_set(offset) then begin
   ; Rough offset between im1 and im2
   rough_dx   = double(offset(0))
   rough_dy   = double(offset(1))
   xmatch1    = dblarr(n_line2)
   ymatch1    = dblarr(n_line2)
   rough_dist = dblarr(n_line2)  ; Residual distance after rough offset
   for k=0,n_line2-1 do begin
      dist = sqrt((xlist2(k)-xlist1-rough_dx)^2 + $
                  (ylist2(k)-ylist1-rough_dy)^2)
      ; Sort in order of increasing distance
      order         = sort(dist)
      best          = order(0)
      xmatch1(k)    = xlist1(best)
      ymatch1(k)    = ylist1(best)
      rough_dist(k) = dist(best)
   endfor
endif else begin
   ; Start at zero offset and step around in x and y to try and
   ; improve the match up.  If the best matching offset is at
   ; the edge of the trial range, repeat around that position.
   xcent = 0
   ycent = 0
   done  = 0
   while not(done) do begin
      ; Find the offset that minimizes the median distance 
      ; between point on im2 and the nearest point on im1
      nside    = 2
      xoffset  = dindgen(2*nside+1) - nside + xcent
      yoffset  = dindgen(2*nside+1) - nside + ycent
      all_dist = dblarr(2*nside+1,2*nside+1,n_line2)
      for i=0,2*nside do begin
         for j=0,2*nside do begin
            xmatch1   = dblarr(n_line2)
            ymatch1   = dblarr(n_line2)
            dist_list = dblarr(n_line2)
            temp_dx = xoffset(i)
            temp_dy = yoffset(j)
            for k=0,n_line2-1 do begin
               dist = sqrt((xlist2(k)-xlist1-temp_dx)^2 + $
                           (ylist2(k)-ylist1-temp_dy)^2)
               ; Sort in order of increasing distance
               order        = sort(dist)
               best         = order(0)
               xmatch1(k)   = xlist1(best)
               ymatch1(k)   = ylist1(best)
               dist_list(k) = dist(best)
            endfor
            all_dist(i,j,*) = dist_list 
         endfor
      endfor
      ; Determine the best offset
      med_dist   = median(all_dist,dim=3)
      med_order  = sort(med_dist)
      best_index = med_order(0)
      best_ij    = array_indices(med_dist,best_index)
      ; If the best offset is at the edge of the trial range,
      ; repeat the process centered at that point
      if (min(best_ij) eq 0 or max(best_ij) eq 2*nside+1) then begin
         xcent = xoffset(best_ij(0))
         ycent = yoffset(best_ij(1))
      endif else begin
         rough_dx   = xoffset(best_ij(0))
         rough_dy   = yoffset(best_ij(1))
         rough_dist = all_dist(best_ij(0),best_ij(1),*)  ; Residual distance
         done       = 1
      endelse
   endwhile
endelse

;;; Reject lines where the best nearest distance is outside the norm
xgood1      = xmatch1
ygood1      = ymatch1
xgood2      = xlist2
ygood2      = ylist2
good_dist   = rough_dist
done_reject = 0
while not(done_reject) do begin
   sdev    = robust_sigma(good_dist)
   keep_ls = where(abs(good_dist-median(good_dist)) le 2*sdev,n_keep,$
                   complement=reject_ls,ncomplement=n_reject)
   if (n_reject ne 0) then begin
      xgood1    = xgood1(keep_ls)
      ygood1    = ygood1(keep_ls)
      xgood2    = xgood2(keep_ls)
      ygood2    = ygood2(keep_ls)
      good_dist = good_dist(keep_ls)
   endif else begin
      n_good      = n_keep
      done_reject = 1
   endelse
endwhile
delta_x = xgood2 - xgood1
delta_y = ygood2 - ygood1

;;; Surface polynomial orders
if keyword_set(nx) then xpolyorder = nx else xpolyorder = 1
if keyword_set(ny) then ypolyorder = ny else ypolyorder = 1

;;; Do a surface fit of x1 as a fn of x2 and y2
xfit_done  = 0
xdata      = dblarr(3,n_good)
xdata(0,*) = xgood2
xdata(1,*) = ygood2
xdata(2,*) = xgood1
n_xfit     = n_good
while not(xfit_done) do  begin
   x1fit     = sfit(xdata,xpolyorder,/irregular,kx=xterms,/max_degree)
   med_xerr = median(xdata(2,*)-x1fit)
   sdev_x   = stddev(xdata(2,*)-x1fit)
   keep_ls   = where(abs(xdata(2,*)-x1fit-med_xerr) le 3*sdev_x,n_keep)
   if (n_keep eq n_xfit) then begin
      xfit_done = 1
   endif else begin
      xdata = xdata(*,keep_ls)
      n_xfit = n_keep
   endelse
endwhile

;;; Do a surface fit of y1 as a fn of x2 and y2
yfit_done  = 0
ydata      = dblarr(3,n_good)
ydata(0,*) = xgood2
ydata(1,*) = ygood2
ydata(2,*) = ygood1
n_yfit     = n_good
while not(yfit_done) do  begin
   y1fit     = sfit(ydata,ypolyorder,/irregular,kx=yterms,/max_degree)
   med_yerr = median(ydata(2,*)-y1fit)
   sdev_y   = stddev(ydata(2,*)-y1fit)
   keep_ls   = where(abs(ydata(2,*)-y1fit-med_yerr) le 3*sdev_y,n_keep)
   if (n_keep eq n_yfit) then begin
      yfit_done = 1
   endif else begin
      ydata = ydata(*,keep_ls)
      n_yfit = n_keep
   endelse
endwhile

;;; Construct p and q arrays to be input into POLY_2D.  Thesse
;;; must be n+1 x n+1 arrays, where n is the larger of xpolyorder
;;; and ypolyorder.  This assumes that the /MAX_DEGREE keyword
;;; was set in SFIT above.
max_polyorder = xpolyorder > ypolyorder
p = fltarr(max_polyorder+1,max_polyorder+1)
k = 0
for i=0,xpolyorder do begin      & $
   for j=0,xpolyorder-i do begin & $
      p(j,i) = xterms(k)         & $
      k = k+1                    & $
   endfor                        & $
endfor
q = fltarr(max_polyorder+1,max_polyorder+1)
k = 0
for i=0,ypolyorder do begin      & $
   for j=0,ypolyorder-i do begin & $
      q(j,i) = yterms(k)         & $
      k = k+1                    & $
   endfor                        & $
endfor

;;; Coordinates in im1 corresponding to (0,0) and (xsz-1,ysz-1) in im2
;;; These can be input into im_rot (linear transformations only).
x = [0d0,double(xsz-1)]
y = [0d0,double(ysz-1)]
newxcorner = dblarr(2)
newycorner = dblarr(2)
for i=0,max_polyorder do begin
   for j=0,max_polyorder do begin
      newxcorner = newxcorner + p(j,i)*(x^i)*(y^j)
      newycorner = newycorner + q(j,i)*(x^i)*(y^j)
   endfor
endfor
coord1 = [x,y]
coord2 = [newxcorner,newycorner]

;;; Swap x and y polynomial coefficients and
;;; corner coordinates
tmp_p  = p
tmp_q  = q
p      = rotate(tmp_q,4)
q      = rotate(tmp_p,4)
coord1 = [y,x]
coord2 = [newycorner,newxcorner]

return
end
