Pro esi_slitgrid,stdfile,slitgrid,orderlist,flatfile=flatfile

;+--------------------------------------------------------------------------
;
; ESI_SLITGRID      9/2010
;
; Trace orders in a bright standard star exposure to create a map of slit
; positions.  The slit position at a pixel is the x-intercept of the
; curve of constant position along the slit that would pass through
; the center of that pixel.
;
; Standard star should be at low airmass (less than 1.1)
;
; Assumes illuminated region is 2048x4096 (1x1 binning)
;
; INPUTS
;         stdfile   - String, FITS filename of standard star exposure
;
; KEYWORDS
;         flatfile  - String, FITS filename of flatfield exposure
;                     (non-diffused).  Used to identify the edges
;                     of each order.
;
; OUTPUTS
;         slitgrid  - Name of FITS file to which to write slit grid
;         orderlist - Name of ASCII file to which to write list of
;                     order centers and edges.  Output only if the
;                     flatfile keyword is set.
;
; HISTORY
;         Adapted from MAGE_SLITGRID 9/11/2010
;         Modified way in which order centers are calculated 9/16/2011
;---------------------------------------------------------------------------

if (n_params() lt 2) then begin
 print,'CALLING SEQUENCE: '
 print,'   esi_slitgrid,stdfile,slitgrid,orderlist,flatfile=flatfile'
 return
endif

device,decomp=0
loadct,39

;;; Overscan sizes
prescan_xsz  = 24
overscan_xsz = 160

;;; Rough size of slit, including a buffer
slitsz = 175  ; 1x1 binning

;;; Read in standard star exposure
std  = esi_subbias(readfits(stdfile,hstd))
sz   = size(std)
xsz  = sz(1)
ysz  = sz(2)
illum_xsz = xsz - prescan_xsz - overscan_xsz

;;; Arrays of x and y values
xarr = fltarr(xsz,ysz)
yarr = fltarr(xsz,ysz)
for i=0,xsz-1 do xarr(i,*) = i
for i=0,ysz-1 do yarr(*,i) = i

;;; Median cut through the array
cut_ymin = ysz/2.
cut_ymax = cut_ymin + 20
cut_ymid = 0.5*(cut_ymin+cut_ymax)
cut      = median(std(*,cut_ymin:cut_ymax),dim=2)

;;; Locate peaks.  Avoid edges of illuminated region.
first_peak = 1
med_cut    = median(cut)
cut_xmin   = prescan_xsz + 10
cut_xmax   = prescan_xsz + illum_xsz - 11
for j=cut_xmin,cut_xmax do begin
   med_xmin = (0 > (j-40)) < (prescan_xsz+illum_xsz-1-80)
   med_xmax = med_xmin + 80
   if (cut(j) eq max(cut(j-10:j+10)) and $
       cut(j) gt median(cut(med_xmin:med_xmax)) + $
                    7*robust_sigma(cut(med_xmin:med_xmax)) and $
       cut(j) lt 60000. and $          ; Avoid saturated lines
       cut(j) ne cut(j-1)) then begin  ; Avoid double-counting
      if (first_peak) then begin
         peak_list  = j
         first_peak = 0
      endif else begin
         peak_list  = [peak_list,j]
      endelse
   endif
endfor
n_peak = n_elements(peak_list)

;;; Plot peak locations
window,9,xs=1000,ys=500,title='Peak Locations'
!p.multi = [0,1,1]
plot,cut,psym=10
for i=0,n_peak-1 do oplot,replicate(peak_list(i),2),[-1e6,1e6],color=250

;;; Illuminated regions
xrange = [0,prescan_xsz+illum_xsz-1]
yrange = [0,ysz-1]

;;; Master slit grid.  Flag non-illuminated regions.
slit = fltarr(xsz,ysz) - 1000

;;; X-intercept of each trace
x_intercept = fltarr(n_peak)

;;; Y-values where trace has a kink - 10 orders
y_break = [649,573,469,343,175,0,0,0,0,0]

;;; Trace, fit, and compute slit grid in the vicinity of each peak.
print,'Tracing...'
window,10,xs=1000,ys=800,title='Traces'
!p.multi = [0,ceil(sqrt(n_peak)),ceil(sqrt(n_peak))]
for j=0,n_peak-1 do begin
   ; Avoid under-illuminated corners
   xstart = peak_list(j)
   ystart = cut_ymid
   traceline,std,xstart,ystart,xtr,ytr,/ydir,merit=merit,$
             median=3,range=yrange,bound=xrange,nside=10,/noplot,/bygauss
   ; Reject bad pixels
   keep_ls = where(merit gt 0 and ytr ge y_break(j),n_keep)
   xtr     = xtr(keep_ls)
   ytr     = ytr(keep_ls)
   merit   = merit(keep_ls)
   ; Use a lower-order polynomial for the preliminary fit.
   npoly_prelim  = 4
   temp_trcoeffs = robust_poly_fit(ytr,xtr,npoly_prelim,xfit,xerr)
   ; Reject outliers and refit to get uncertainties in
   ; coefficients.  User a higher-order polynomial for longer orders.
   ls = where(abs(xtr-xfit) le ((3*xerr) > 1.0))
   ylength = max(ytr(ls))-min(ytr(ls))
   npoly_trace = 4
   ;if (ylength gt 2000) then npoly_trace = 5
   trcoeffs = poly_fit(ytr(ls),xtr(ls),npoly_trace,$
                       sigma=sig,yerror=xerr,yfit=xfit)
   x_intercept(j) = trcoeffs(0)
   ; Plot residuals
   plot,xtr(ls)-xfit,ytr(ls),psym=3,xs=3,/ystyle,xr=[-1,1],$
        title=strtrim(j,2),charsize=1.5
   oplot,[0,0],[-1e6,1e6],linestyle=2
   ; Create a temporary 2D slit array assuming all loci of constant
   ; slit position run parallel to this trace
   temp_slit = xarr
   for k=1,npoly_trace do temp_slit = temp_slit - trcoeffs(k)*(yarr^k)
   ; Mask pre and overscan regions
   temp_slit(0:prescan_xsz-1,*) = -1000
   temp_slit(xsz-overscan_xsz:xsz-1,*) = -1000
   ; Extract the region around this trace and save in the master
   ; slitgrid.  Assume the std is at the center of the slit, and
   ; include a margin at the edge of the slit.  Use wider margins for
   ; bluer orders.  Only save the blue end of the reddest order.
   if (j eq n_peak-1) then separation = (peak_list(j)-peak_list(j-1)) $
      else separation = (peak_list(j+1)-peak_list(j))
   order_ls = where(temp_slit ge trcoeffs(0) - separation/2. and $
                    temp_slit le trcoeffs(0) + separation/2.)
   slit(order_ls) = temp_slit(order_ls)
endfor

;;; Write slitgrid to file
sxaddpar,hstd,'BZERO',0.0
writefits,slitgrid,slit,hstd

;;; Identify order edges (optional)
if keyword_set(flatfile) then begin
   flat = esi_subbias(readfits(flatfile,hflat))
   window,11,xs=1000,ys=500
   !p.multi   = [0,1,1]
   print,'Left-click once on each edge of each order, left then right.'
   print,'Right-click when done with each order.'
   slit_slice = slit(0:illum_xsz-1,1800:2000)
   flat_slice = flat(0:illum_xsz-1,1800:2000)
   orderedge_list = fltarr(2*n_peak)
   for i=0,n_peak-1 do begin
      x_min = x_intercept(i) - (2./3)*slitsz
      x_max = x_intercept(i) + (2./3)*slitsz
      y_min = 0
      plot_ls = where(slit_slice ge x_min and slit_slice le x_max)
      y_max = cutvalue(flat_slice(plot_ls),99.9)
      plot,slit_slice(plot_ls),flat_slice(plot_ls),psym=3,xr=[x_min,x_max],/xs,$
           yr=[y_min,y_max]
      getpoints,temp_orderedge,/plotx,color=150
      orderedge_list(2*i:2*i+1) = temp_orderedge
   endfor
   
   ;;; Compute center of each order
   n_orderedge = n_elements(orderedge_list)
   if (n_orderedge mod 2 ne 0) then begin
      print,'ESI_SLITGRID: Error - There should be an even number of order' 
      print,'               edges marked (two for each order).  Exiting without'
      print,'               writing order list.'
      return
   endif
   n_order        = n_orderedge / 2
   leftedge_list  = orderedge_list(2*lindgen(n_order))
   rightedge_list = orderedge_list(1+2*lindgen(n_order))
   ;center_list    = (leftedge_list + rightedge_list) / 2.
   ; Fit a smooth function to the order midpoints
   midpoint_list  = (leftedge_list + rightedge_list) / 2.
   center_polyorder = floor(n_order/2.) < 5
   ordernum         = findgen(n_order)
   center_pfit      = robust_poly_fit(ordernum,midpoint_list,center_polyorder)
   center_list      = 0.*midpoint_list
   for i=0,center_polyorder do center_list = center_list + $
                                             center_pfit(i)*(ordernum^i)
   
   ;;; Write order info to file
   print,'Writing order info to file...'
   openw,unit,orderlist,/get_lun
   printf,unit,'#  CCD   Order    Center   Left edge  Right edge'
   for i=0,n_order-1 do $
      printf,unit,'ESI',i+1,center_list(i),leftedge_list(i),$
                  rightedge_list(i),format='(a6,i6,3(f12.3))'
   free_lun,unit
endif

; Restore plot settings
cleanplot,/silent

return
end
