Function esi_subbias,raw

;+---------------------------------------------------------------------
;
; ESI_SUBBIAS     9/2010 GDB
; 
; Subtract the bias from a raw ESI frame by computing the mean bias
; level in the overscan region.
;
; INPUTS
;          raw      - Raw 2D array, can be a named variable or the name
;                     of a FITS file.
;
; OUTPUTS
;          <result> - 2D bias-subtracted array.
;
; HISTORY
;          Written 1/18/2008 GDB
;---------------------------------------------------------------------

if (n_params() eq 0) then begin
   print,'CALLING SEQUENCE: result = esi_subbias(raw)'
   return,-1
endif

; Read in FITS file, if necessary
if (size(raw,/tname) eq 'STRING') then begin
   checkexists = findfile(raw,count=nfiles)
   if (nfiles eq 0) then begin
      print,'ESI_SUBBIAS: Raw file not found.'
      return,-1
   endif else begin
       rawframe = readfits(raw,hrawframe)
       fromfitsfile = 1 
   endelse
endif else rawframe = raw

; Arrays of x-values covering pre-scan, over-scan, and illuminated
; regions (assumes 2232 x 4096 frame)
prescan_x1  =    0 + lindgen(12)
prescan_x2  =   12 + lindgen(12) 
illum_x1    =   24 + lindgen(1024)
illum_x2    = 1048 + lindgen(1024)
overscan_x1 = 2072 + lindgen(80)
overscan_x2 = 2152 + lindgen(80)

; Size of illuminated region + overscan
sz  = size(rawframe)
xsz = sz(1)
ysz = sz(2)
if (xsz ne 2232) then begin
   if (xsz eq 2048) then begin
      ; Special case for incorrect CCD readout settings: Insert fake
      ; pre- and overscan regions with approximate bias level.  Bias
      ; will vary between exposures, and may have both a tilt and
      ; substructure in the y-direction, but these values should be
      ; accurate to within a few counts.
      print,'Using approximate bias levels'
      approx_bias1 = 1003.
      approx_bias2 = 997.
      newframe = fltarr(2232,4096)
      newframe(24:2071,*)     = rawframe
      newframe(prescan_x1,*)  = approx_bias1
      newframe(prescan_x2,*)  = approx_bias2
      newframe(overscan_x1,*) = approx_bias1
      newframe(overscan_x2,*) = approx_bias2
      rawframe = newframe
   endif else begin
      print,'Unexpected x size'
      return,-1
   endelse
endif

; Bias-subtracted frame
subframe = 1.*rawframe  ; Promote to floating point

; Measure the mean bias level across the overscan region in x and
; subtract from each row.  Do bad pixel rejection.
for i=0,ysz-1 do begin
   ; Amp 1
   tmp_oscan1    = subframe(overscan_x1,i)
   med_level1    = median(tmp_oscan1)
   sig_level1    = robust_sigma(tmp_oscan1)
   good_ls1      = where(abs(tmp_oscan1-med_level1) le 4*sig_level1)
   mean_level1   = mean(tmp_oscan1(good_ls1))
   subframe(prescan_x1,i)  = subframe(prescan_x1,i)  - mean_level1
   subframe(illum_x1,i)    = subframe(illum_x1,i)    - mean_level1
   subframe(overscan_x1,i) = subframe(overscan_x1,i) - mean_level1
   ; Amp 2
   tmp_oscan2    = subframe(overscan_x2,i)
   med_level2    = median(tmp_oscan2)
   sig_level2    = robust_sigma(tmp_oscan2)
   good_ls2      = where(abs(tmp_oscan2-med_level2) le 4*sig_level2)
   mean_level2   = mean(tmp_oscan2(good_ls2))
   subframe(prescan_x2,i)  = subframe(prescan_x2,i)  - mean_level2
   subframe(illum_x2,i)    = subframe(illum_x2,i)    - mean_level2
   subframe(overscan_x2,i) = subframe(overscan_x2,i) - mean_level2
endfor

return,subframe
end
