Function im_rot,arr,x0,y0,x1,y1,coord0=coord0,coord1=coord1,$
                trans=trans

;+------------------------------------------------------------------------
;
; IM_ROT         3/2007 
;
; Align a 2D image to a second image using ROT.  Computes the rotation
; angle, magnification, and x,y coordinates of the center of rotation
; based on the x,y coordinates of two points in the input and output
; images.
;
; Assumes the input image and the image to which it is to be matched
; have the same dimensions.
;
; INPUTS
;        arr    - 2D image to be aligned
;        *** The following inputs must be supplied unless the coord0 ***
;        *** and coord1 keywords are set ***
;        x0,y0  - 2-element vectors giving the x and y-coordinates of
;                 two points, a and b, in the input image [x0a,x0b],
;                 [y0a,y0b]
;        x1,y1  - 2-element vectors giving the desired x and
;                 y-coordinates of the same two points in the output
;                 image [x1a,x1b], [y1a,y1b].
;
; KEYWORDS
;        coord0   - Coordinates in the input image, [x0a,x0b,y0a,y0b]
;                   If set, overides x0,y0 inputs.  Must set both
;                   coord0 and coord1 keywords.
;        coord1   - Coordinates in the output image, [x1a,x1b,y1a,y1b]
;                   If set, overides x1,y1 inputs.  Must set both
;                   coord0 and coord1 keywords.
;        trans - Parameters sent to ROT [angle,mag,xc,yc]
;
; OUTPUTS
;        <output> - Rotated, magnified, and translated image, with
;                   same dimensions as the input image.  Missing
;                   elements are set to 0.
;
; HISTORY
;        Written 3/19/2007 GDB
;        COORD1 and COORD2 keywords added 7/28/2007 GDB
;-------------------------------------------------------------------------

if not(n_params() eq 5 or (n_params() eq 1 and $
       keyword_set(coord0) and keyword_set(coord1))) then begin
   print,'CALLING SEQUENCE:'
   print,'       new_image = im_rot(arr,x0,y0,x1,y1,trans=trans)'
   print,' OR,   new_image = im_rot(arr,coord0=coord0,coord1=coord1,trans=trans)'
   return,-1
endif

; Size of input image.
sz  = size(arr)
if (sz(0) ne 2) then begin
   print,'Input image must be a 2D array.'
   return,-1
endif
xsz = double(sz(1))
ysz = double(sz(2))

; *** NOTE: Below, the input image is referred to as image 2.
; *** The image to which the input image is to be matched is
; *** referred to as image 1.
;
; Convention: A, B, and O are points on image 1, where
; O is at the center of the array.
;
; A', B', and O' are the corresponding points on image 2, 
; i.e., O' is the point on image 2 that will be at the
; center of the rotated/translated/magnified image.

if (keyword_set(coord0) and keyword_set(coord1)) then begin

   ; Coordinates of points on image 1
   xa1 = double(coord1(0))  &  ya1 = double(coord1(2))
   xb1 = double(coord1(1))  &  yb1 = double(coord1(3))
   
   ; Coordinates of matching points on image 2 (inpt image)
   xa2 = double(coord0(0))  &  ya2 = double(coord0(2))
   xb2 = double(coord0(1))  &  yb2 = double(coord0(3))   

endif else begin

   ; Coordinates of points on image 1
   xa1 = double(x1(0))  &  ya1 = double(y1(0))
   xb1 = double(x1(1))  &  yb1 = double(y1(1))
   
   ; Coordinates of matching points on image 2 (inpt image)
   xa2 = double(x0(0))  &  ya2 = double(y0(0))
   xb2 = double(x0(1))  &  yb2 = double(y0(1))

endelse

; Center of the array on image 1
xo1 = (xsz-1d0)/2d0
yo1 = (ysz-1d0)/2d0

; Distance between A and B on image 1
dx_ab1 = xb1 - xa1
dy_ab1 = yb1 - ya1
ab1    = sqrt(dx_ab1^2 + dy_ab1^2)

; Distance between A and O on image 1
dx_ao1 = xo1 - xa1
dy_ao1 = yo1 - ya1
ao1    = sqrt(dx_ao1^2 + dy_ao1^2)

; Angle between the line connecting A and B and vertical
; lines though A on image 1
angle_ab1 = atan(dx_ab1/dy_ab1)  ;***********

; Angle between the line connecting A and O and a vertical line
; through A on image 1
angle_ao1 = atan(dx_ao1/dy_ao1)

; Angle between the line connecting A and B and the line connecting A
; and the center on image 1
angle_Q = angle_ab1 - angle_ao1

; Distance between A' and B' on image 2
dx_ab2 = xb2 - xa2
dy_ab2 = yb2 - ya2
ab2    = sqrt(dx_ab2^2 + dy_ab2^2)

; Distance between A' and O' on image 2
ao2 = ao1*(ab2/ab1)

; Angle between the line connecting A' and B' and vertical
; lines though A' on image 2
angle_ab2 = atan(dx_ab2/dy_ab2)

; Angle between the line connecting A' and O' and a vertical line
; through A' on image 2
angle_ao2 = angle_ab2 - angle_Q

; Coordinates of O' on image 2
dx_ao2 = ao2*sin(angle_ao2)
dy_ao2 = ao2*cos(angle_ao2)
xo2    = xa2 - dx_ao2  ; *** These signs can either be (+) or (-).
yo2    = ya2 - dy_ao2  ; *** Must be able to tell which one!

; Rotation clockwise to match image 2 to image 1
angle_rad = angle_ab1 - angle_ab2
angle_deg = angle_rad * (360d0/(2d0*!dpi))

; Magnification to match image 2 to image 1
mag = ab1 / ab2

; Rotate/magnify/translate image 2
new_arr = rot(arr,angle_deg,mag,xo2,yo2,/interp,missing=0)

; Transformation parameters
trans = [angle_deg,mag,xo2,yo2]

return,new_arr
end
